==> allensdk-2.13.0-py3-none-any.whl: allensdk/brain_observatory/behavior/data_objects/eye_tracking/rig_geometry.py <==
import warnings

import numpy as np
import pandas as pd
from typing import Optional

import pynwb
from pynwb import NWBFile

from allensdk.brain_observatory.behavior.data_objects import DataObject
from allensdk.brain_observatory.behavior.data_objects.base \
    .readable_interfaces import \
    LimsReadableInterface, JsonReadableInterface, NwbReadableInterface
from allensdk.brain_observatory.behavior.data_objects.base\
    .writable_interfaces import \
    NwbWritableInterface
from allensdk.brain_observatory.behavior.schemas import \
    OphysEyeTrackingRigMetadataSchema
from allensdk.brain_observatory.nwb import load_pynwb_extension
from allensdk.internal.api import PostgresQueryMixin


class Coordinates:
    """Represents coordinates in 3d space"""
    def __init__(self, x: float, y: float, z: float):
        self._x = x
        self._y = y
        self._z = z

    @property
    def x(self):
        return self._x

    @property
    def y(self):
        return self._y

    @property
    def z(self):
        return self._z

    def __iter__(self):
        yield self._x
        yield self._y
        yield self._z

    def __eq__(self, other):
        if type(other) not in (type(self), list):
            raise ValueError(f'Do not know how to compare with type '
                             f'{type(other)}')
        if isinstance(other, list):
            return self._x == other[0] and \
                   self._y == other[1] and \
                   self._z == other[2]
        else:
            return self._x == other.x and \
                   self._y == other.y and \
                   self._z == other.z

    def __str__(self):
        return f'[{self._x}, {self._y}, {self._z}]'


class RigGeometry(DataObject, LimsReadableInterface, JsonReadableInterface,
                  NwbReadableInterface, NwbWritableInterface):
    def __init__(self, equipment: str,
                 monitor_position_mm: Coordinates,
                 monitor_rotation_deg: Coordinates,
                 camera_position_mm: Coordinates,
                 camera_rotation_deg: Coordinates,
                 led_position: Coordinates):
        super().__init__(name='rig_geometry', value=self)
        self._monitor_position_mm = monitor_position_mm
        self._monitor_rotation_deg = monitor_rotation_deg
        self._camera_position_mm = camera_position_mm
        self._camera_rotation_deg = camera_rotation_deg
        self._led_position = led_position
        self._equipment = equipment

    @property
    def monitor_position_mm(self):
        return self._monitor_position_mm

    @property
    def monitor_rotation_deg(self):
        return self._monitor_rotation_deg

    @property
    def camera_position_mm(self):
        return self._camera_position_mm

    @property
    def camera_rotation_deg(self):
        return self._camera_rotation_deg

    @property
    def led_position(self):
        return self._led_position

    @property
    def equipment(self):
        return self._equipment

    def to_nwb(self, nwbfile: NWBFile) -> NWBFile:
        eye_tracking_rig_mod = pynwb.ProcessingModule(
            name='eye_tracking_rig_metadata',
            description='Eye tracking rig metadata module')

        nwb_extension = load_pynwb_extension(
            OphysEyeTrackingRigMetadataSchema, 'ndx-aibs-behavior-ophys')

        rig_metadata = nwb_extension(
            name="eye_tracking_rig_metadata",
            equipment=self._equipment,
            monitor_position=list(self._monitor_position_mm),
            monitor_position__unit_of_measurement="mm",
            camera_position=list(self._camera_position_mm),
            camera_position__unit_of_measurement="mm",
            led_position=list(self._led_position),
            led_position__unit_of_measurement="mm",
            monitor_rotation=list(self._monitor_rotation_deg),
            monitor_rotation__unit_of_measurement="deg",
            camera_rotation=list(self._camera_rotation_deg),
            camera_rotation__unit_of_measurement="deg"
        )

        eye_tracking_rig_mod.add_data_interface(rig_metadata)
        nwbfile.add_processing_module(eye_tracking_rig_mod)
        return nwbfile

    @classmethod
    def from_nwb(cls, nwbfile: NWBFile) -> Optional["RigGeometry"]:
        try:
            et_mod = \
                nwbfile.get_processing_module("eye_tracking_rig_metadata")
        except KeyError as e:
            warnings.warn("This ophys session "
                          f"'{int(nwbfile.identifier)}' has no eye "
                          f"tracking rig metadata. (NWB error: {e})")
            return None

        meta = et_mod.get_data_interface("eye_tracking_rig_metadata")

        monitor_position = meta.monitor_position[:]
        monitor_position = (monitor_position.tolist()
                            if isinstance(monitor_position, np.ndarray)
                            else monitor_position)

        monitor_rotation = meta.monitor_rotation[:]
        monitor_rotation = (monitor_rotation.tolist()
                            if isinstance(monitor_rotation, np.ndarray)
                            else monitor_rotation)

        camera_position = meta.camera_position[:]
        camera_position = (camera_position.tolist()
                           if isinstance(camera_position, np.ndarray)
                           else camera_position)

        camera_rotation = meta.camera_rotation[:]
        camera_rotation = (camera_rotation.tolist()
                           if isinstance(camera_rotation, np.ndarray)
                           else camera_rotation)

        led_position = meta.led_position[:]
        led_position = (led_position.tolist()
                        if isinstance(led_position, np.ndarray)
                        else led_position)

        return RigGeometry(
            equipment=meta.equipment,
            monitor_position_mm=Coordinates(*monitor_position),
            camera_position_mm=Coordinates(*camera_position),
            led_position=Coordinates(*led_position),
            monitor_rotation_deg=Coordinates(*monitor_rotation),
            camera_rotation_deg=Coordinates(*camera_rotation)
        )

    @classmethod
    def from_json(cls, dict_repr: dict) -> "RigGeometry":
        rg = dict_repr['eye_tracking_rig_geometry']
        return RigGeometry(
            equipment=rg['equipment'],
            monitor_position_mm=Coordinates(*rg['monitor_position_mm']),
            monitor_rotation_deg=Coordinates(*rg['monitor_rotation_deg']),
            camera_position_mm=Coordinates(*rg['camera_position_mm']),
            camera_rotation_deg=Coordinates(*rg['camera_rotation_deg']),
            led_position=Coordinates(*rg['led_position'])
        )

    @classmethod
    def from_lims(cls, ophys_experiment_id: int,
                  lims_db: PostgresQueryMixin) -> Optional["RigGeometry"]:
        query = f'''
            SELECT oec.*, oect.name as config_type, equipment.name as 
            equipment_name
            FROM ophys_sessions os
            JOIN observatory_experiment_configs oec ON oec.equipment_id = 
            os.equipment_id
            JOIN observatory_experiment_config_types oect ON oect.id = 
            oec.observatory_experiment_config_type_id
            JOIN ophys_experiments oe ON oe.ophys_session_id = os.id
            JOIN equipment ON equipment.id = oec.equipment_id
            WHERE oe.id = {ophys_experiment_id} AND 
                oec.active_date <= os.date_of_acquisition AND
                oect.name IN ('eye camera position', 'led position', 'screen position')
        '''  # noqa E501
        # Get the raw data
        rig_geometry = pd.read_sql(query, lims_db.get_connection())

        if rig_geometry.empty:
            # There is no rig geometry for this experiment
            return None

        # Map the config types to new names
        rig_geometry_config_type_map = {
            'eye camera position': 'camera',
            'screen position': 'monitor',
            'led position': 'led'
        }
        rig_geometry['config_type'] = rig_geometry['config_type'] \
            .map(rig_geometry_config_type_map)

        rig_geometry = cls._select_most_recent_geometry(
            rig_geometry=rig_geometry)

        # Construct dictionary for positions
        position = rig_geometry[['center_x_mm', 'center_y_mm', 'center_z_mm']]
        position.index = [
            f'{v}_position_mm' if v != 'led'
            else f'{v}_position' for v in position.index]
        position = position.to_dict(orient='index')
        position = {
            config_type:
                Coordinates(
                    values['center_x_mm'],
                    values['center_y_mm'],
                    values['center_z_mm'])
            for config_type, values in position.items()
        }

        # Construct dictionary for rotations
        rotation = rig_geometry[['rotation_x_deg', 'rotation_y_deg',
                                 'rotation_z_deg']]
        rotation = rotation[rotation.index != 'led']
        rotation.index = [f'{v}_rotation_deg' for v in rotation.index]
        rotation = rotation.to_dict(orient='index')
        rotation = {
            config_type:
                Coordinates(
                    values['rotation_x_deg'],
                    values['rotation_y_deg'],
                    values['rotation_z_deg']
                )
                for config_type, values in rotation.items()
        }

        # Combine the dictionaries
        rig_geometry = {
            **position,
            **rotation,
            'equipment': rig_geometry['equipment_name'].iloc[0]
        }
        return RigGeometry(**rig_geometry)

    @staticmethod
    def _select_most_recent_geometry(rig_geometry: pd.DataFrame):
        """There can be multiple geometry entries in LIMS for a rig.
        Select most recent one.

        Parameters
        ----------
        rig_geometry
            Table of geometries for rig as returned by LIMS

         Notes
         ----------
         The geometries in rig_geometry are assumed to precede the
         date_of_acquisition of the session
        (only relevant for retrieving from LIMS)
        """
        rig_geometry = rig_geometry.sort_values('active_date', ascending=False)
        rig_geometry = rig_geometry.groupby('config_type') \
            .apply(lambda x: x.iloc[0])
        return rig_geometry

==> allensdk-2.16.1-py3-none-any.whl: allensdk/brain_observatory/behavior/data_objects/eye_tracking/rig_geometry.py <==
import warnings

import numpy as np
import pandas as pd
from typing import Optional

import pynwb
from pynwb import NWBFile

from allensdk.core import DataObject
from allensdk.core import \
    LimsReadableInterface, JsonReadableInterface, NwbReadableInterface
from allensdk.core import \
    NwbWritableInterface
from allensdk.brain_observatory.behavior.schemas import \
    OphysEyeTrackingRigMetadataSchema
from allensdk.brain_observatory.nwb import load_pynwb_extension
from allensdk.internal.api import PostgresQueryMixin


class Coordinates:
    """Represents coordinates in 3d space"""
    def __init__(self, x: float, y: float, z: float):
        self._x = x
        self._y = y
        self._z = z

    @property
    def x(self):
        return self._x

    @property
    def y(self):
        return self._y

    @property
    def z(self):
        return self._z

    def __iter__(self):
        yield self._x
        yield self._y
        yield self._z

    def __eq__(self, other):
        if type(other) not in (type(self), list):
            raise ValueError(f'Do not know how to compare with type '
                             f'{type(other)}')
        if isinstance(other, list):
            return self._x == other[0] and \
                   self._y == other[1] and \
                   self._z == other[2]
        else:
            return self._x == other.x and \
                   self._y == other.y and \
                   self._z == other.z

    def __str__(self):
        return f'[{self._x}, {self._y}, {self._z}]'


class RigGeometry(DataObject, LimsReadableInterface, JsonReadableInterface,
                  NwbReadableInterface, NwbWritableInterface):
    def __init__(self, equipment: str,
                 monitor_position_mm: Coordinates,
                 monitor_rotation_deg: Coordinates,
                 camera_position_mm: Coordinates,
                 camera_rotation_deg: Coordinates,
                 led_position: Coordinates):
        super().__init__(name='rig_geometry', value=None, is_value_self=True)
        self._monitor_position_mm = monitor_position_mm
        self._monitor_rotation_deg = monitor_rotation_deg
        self._camera_position_mm = camera_position_mm
        self._camera_rotation_deg = camera_rotation_deg
        self._led_position = led_position
        self._equipment = equipment

    @property
    def monitor_position_mm(self):
        return self._monitor_position_mm

    @property
    def monitor_rotation_deg(self):
        return self._monitor_rotation_deg

    @property
    def camera_position_mm(self):
        return self._camera_position_mm

    @property
    def camera_rotation_deg(self):
        return self._camera_rotation_deg

    @property
    def led_position(self):
        return self._led_position

    @property
    def equipment(self):
        return self._equipment

    def to_nwb(self, nwbfile: NWBFile) -> NWBFile:
        eye_tracking_rig_mod = pynwb.ProcessingModule(
            name='eye_tracking_rig_metadata',
            description='Eye tracking rig metadata module')

        nwb_extension = load_pynwb_extension(
            OphysEyeTrackingRigMetadataSchema, 'ndx-aibs-behavior-ophys')

        rig_metadata = nwb_extension(
            name="eye_tracking_rig_metadata",
            equipment=self._equipment,
            monitor_position=list(self._monitor_position_mm),
            monitor_position__unit_of_measurement="mm",
            camera_position=list(self._camera_position_mm),
            camera_position__unit_of_measurement="mm",
            led_position=list(self._led_position),
            led_position__unit_of_measurement="mm",
            monitor_rotation=list(self._monitor_rotation_deg),
            monitor_rotation__unit_of_measurement="deg",
            camera_rotation=list(self._camera_rotation_deg),
            camera_rotation__unit_of_measurement="deg"
        )

        eye_tracking_rig_mod.add_data_interface(rig_metadata)
        nwbfile.add_processing_module(eye_tracking_rig_mod)
        return nwbfile

    @classmethod
    def from_nwb(cls, nwbfile: NWBFile) -> Optional["RigGeometry"]:
        try:
            et_mod = \
                nwbfile.get_processing_module("eye_tracking_rig_metadata")
        except KeyError as e:
            warnings.warn("This nwb file with identifier "
                          f"'{int(nwbfile.identifier)}' has no eye "
                          f"tracking rig metadata. (NWB error: {e})")
            return None

        meta = et_mod.get_data_interface("eye_tracking_rig_metadata")

        monitor_position = meta.monitor_position[:]
        monitor_position = (monitor_position.tolist()
                            if isinstance(monitor_position, np.ndarray)
                            else monitor_position)

        monitor_rotation = meta.monitor_rotation[:]
        monitor_rotation = (monitor_rotation.tolist()
                            if isinstance(monitor_rotation, np.ndarray)
                            else monitor_rotation)

        camera_position = meta.camera_position[:]
        camera_position = (camera_position.tolist()
                           if isinstance(camera_position, np.ndarray)
                           else camera_position)

        camera_rotation = meta.camera_rotation[:]
        camera_rotation = (camera_rotation.tolist()
                           if isinstance(camera_rotation, np.ndarray)
                           else camera_rotation)

        led_position = meta.led_position[:]
        led_position = (led_position.tolist()
                        if isinstance(led_position, np.ndarray)
                        else led_position)

        return RigGeometry(
            equipment=meta.equipment,
            monitor_position_mm=Coordinates(*monitor_position),
            camera_position_mm=Coordinates(*camera_position),
            led_position=Coordinates(*led_position),
            monitor_rotation_deg=Coordinates(*monitor_rotation),
            camera_rotation_deg=Coordinates(*camera_rotation)
        )

    @classmethod
    def from_json(cls, dict_repr: dict) -> "RigGeometry":
        rg = dict_repr['eye_tracking_rig_geometry']
        return RigGeometry(
            equipment=rg['equipment'],
            monitor_position_mm=Coordinates(*rg['monitor_position_mm']),
            monitor_rotation_deg=Coordinates(*rg['monitor_rotation_deg']),
            camera_position_mm=Coordinates(*rg['camera_position_mm']),
            camera_rotation_deg=Coordinates(*rg['camera_rotation_deg']),
            led_position=Coordinates(*rg['led_position'])
        )

    @classmethod
    def from_lims(
            cls,
            lims_db: PostgresQueryMixin,
            behavior_session_id: Optional[int] = None,
            ophys_experiment_id: Optional[int] = None
    ) -> Optional["RigGeometry"]:
        if behavior_session_id is None and ophys_experiment_id is None:
            raise ValueError('Must provide either behavior_session_id or '
                             'ophys_experiment_id')
        if behavior_session_id is not None and ophys_experiment_id is not None:
            raise ValueError('Supply either ophys_experiment_id or '
                             'behavior_session_id')
        if ophys_experiment_id is not None:
            query = f'''
                SELECT oec.*, oect.name as config_type, equipment.name as 
                equipment_name
                FROM ophys_experiments oe
                JOIN ophys_sessions os ON oe.ophys_session_id = os.id
                JOIN observatory_experiment_configs oec ON oec.equipment_id = 
                os.equipment_id
                JOIN observatory_experiment_config_types oect ON oect.id = 
                oec.observatory_experiment_config_type_id
                JOIN equipment ON equipment.id = oec.equipment_id
                WHERE oe.id = {ophys_experiment_id} AND 
                    oec.active_date <= os.date_of_acquisition AND
                    oect.name IN ('eye camera position', 'led position', 'screen position')
            '''  # noqa E501
        else:
            query = f'''
                SELECT oec.*, oect.name as config_type, equipment.name as 
                equipment_name
                FROM behavior_sessions bs
                JOIN ophys_sessions os ON os.id = bs.ophys_session_id
                JOIN observatory_experiment_configs oec ON oec.equipment_id = 
                os.equipment_id
                JOIN observatory_experiment_config_types oect ON oect.id = 
                oec.observatory_experiment_config_type_id
                JOIN ophys_experiments oe ON oe.ophys_session_id = os.id
                JOIN equipment ON equipment.id = oec.equipment_id
                WHERE bs.id = {behavior_session_id} AND 
                    oec.active_date <= os.date_of_acquisition AND
                    oect.name IN ('eye camera position', 'led position', 'screen position')
            '''  # noqa E501
        # Get the raw data
        rig_geometry = pd.read_sql(query, lims_db.get_connection())

        if rig_geometry.empty:
            # There is no rig geometry for this experiment
            return None

        # Map the config types to new names
        rig_geometry_config_type_map = {
            'eye camera position': 'camera',
            'screen position': 'monitor',
            'led position': 'led'
        }
        rig_geometry['config_type'] = rig_geometry['config_type'] \
            .map(rig_geometry_config_type_map)

        rig_geometry = cls._select_most_recent_geometry(
            rig_geometry=rig_geometry)

        # Construct dictionary for positions
        position = rig_geometry[['center_x_mm', 'center_y_mm', 'center_z_mm']]
        position.index = [
            f'{v}_position_mm' if v != 'led'
            else f'{v}_position' for v in position.index]
        position = position.to_dict(orient='index')
        position = {
            config_type:
                Coordinates(
                    values['center_x_mm'],
                    values['center_y_mm'],
                    values['center_z_mm'])
            for config_type, values in position.items()
        }

        # Construct dictionary for rotations
        rotation = rig_geometry[['rotation_x_deg', 'rotation_y_deg',
                                 'rotation_z_deg']]
        rotation = rotation[rotation.index != 'led']
        rotation.index = [f'{v}_rotation_deg' for v in rotation.index]
        rotation = rotation.to_dict(orient='index')
        rotation = {
            config_type:
                Coordinates(
                    values['rotation_x_deg'],
                    values['rotation_y_deg'],
                    values['rotation_z_deg']
                )
                for config_type, values in rotation.items()
        }

        # Combine the dictionaries
        rig_geometry = {
            **position,
            **rotation,
            'equipment': rig_geometry['equipment_name'].iloc[0]
        }
        return RigGeometry(**rig_geometry)

    @staticmethod
    def _select_most_recent_geometry(rig_geometry: pd.DataFrame):
        """There can be multiple geometry entries in LIMS for a rig.
        Select most recent one.

        Parameters
        ----------
        rig_geometry
            Table of geometries for rig as returned by LIMS

         Notes
         ----------
         The geometries in rig_geometry are assumed to precede the
         date_of_acquisition of the session
        (only relevant for retrieving from LIMS)
        """
        rig_geometry = rig_geometry.sort_values('active_date', ascending=False)
        rig_geometry = rig_geometry.groupby('config_type') \
            .apply(lambda x: x.iloc[0])
        return rig_geometry
